/* Copyright (C) |Meso|Star> 2016-2017 (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

#ifndef S3DUT_H
#define S3DUT_H

#include <rsys/rsys.h>

/* Library symbol management */
#if defined(S3DUT_SHARED_BUILD) /* Build shared library */
  #define S3DUT_API extern EXPORT_SYM
#elif defined(S3DUT_STATIC_BUILD) /* Use/build static library */
  #define S3DUT_API extern LOCAL_SYM
#else /* Use shared library */
  #define S3DUT_API extern IMPORT_SYM
#endif

/* Helper macro that asserts if the invocation of the s3dut function `Func'
 * returns an error. */
#ifndef NDEBUG
  #define S3DUT(Func) ASSERT(s3dut_##Func == RES_OK)
#else
  #define S3DUT(Func) s3dut_##Func
#endif

struct mem_allocator;
struct s3dut_mesh;

struct s3dut_mesh_data {
  const double* positions;
  const size_t* indices;
  size_t nvertices; /* # vertices */
  size_t nprimitives; /* # primitives */
};

/*******************************************************************************
 * Stard-3DUT API
 ******************************************************************************/
S3DUT_API res_T
s3dut_mesh_ref_get
  (struct s3dut_mesh* mesh);

S3DUT_API res_T
s3dut_mesh_ref_put
  (struct s3dut_mesh* mesh);

S3DUT_API res_T
s3dut_mesh_get_data
  (const struct s3dut_mesh* mesh,
   struct s3dut_mesh_data* data);

/* Create a triangulated UV sphere centered in 0 discretized in `nslices'
 * around the Z axis and `nstacks' along the Z axis. Face vertices are CCW
 * ordered with respect to the sphere center, i.e. they are CW ordered from the
 * outside point of view. */
S3DUT_API res_T
s3dut_create_sphere
  (struct mem_allocator* allocator, /* May be NULL <=> use default allocator */
   const double radius, /* In ]0, INF) */
   const unsigned nslices, /* # subdivisions around Z axis in [3, INF) */
   const unsigned nstacks, /* # subdivisions along Z axis in [2, INF) */
   struct s3dut_mesh** sphere);

/* Create a triangulated cylinder centered in 0 discretized in `nslices' around
 * the Z axis and `nstacks' along the Z axis. The top and the bottom of the
 * cylinder is closed with a triangle fan whose center is on the Z axis. Face
 * vertices are CCW ordered with respect to the cylinder center, i.e. they are
 * CW ordered from the outside
 * point of view. */
S3DUT_API res_T
s3dut_create_cylinder
  (struct mem_allocator* allocator, /* May be NULL <=> use default allocator */
   const double radius, /* In ]0, INF) */
   const double height, /* In ]0, INF) */
   const unsigned nslices, /* # subdivisions around Z axis in [3, INF) */
   const unsigned nstacks, /* # subdivision along Z axis in [1, INF) */
   struct s3dut_mesh** cylinder);

/* Create a triangulated cuboid centered in 0. Face vertices are CCW ordered
 * with respect to the cylinder center, i.e. they are CW ordered from the
 * outside point of view. */
S3DUT_API res_T
s3dut_create_cuboid
  (struct mem_allocator* allocator,
   const double width,
   const double height,
   const double depth,
   struct s3dut_mesh** cuboid);

/* Create a triangulated UV hemisphere oriented wrt Z axis. It discretized in
 * `nslices' around the Z axis and `nstacks' along the Z axis. The back of the
 * hemisphere is not triangulated. Face vertices are CCW ordered with respect
 * to the hemisphere center, i.e. they are CW ordered from the outside point of
 * view. */
S3DUT_API res_T
s3dut_create_hemisphere
  (struct mem_allocator* allocator,
   const double radius,
   const unsigned nslices, /* # subdivisions around Z axis in [2, INF) */
   const unsigned nstacks, /* # subdivisions along Z axis int [2, INF) */
   struct s3dut_mesh** hemisphere);


#endif /* S3DUT_H */

