/* Copyright (C) |Meso|Star> 2016-2017 (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

#include "ssf.h"
#include "ssf_bxdf_c.h"

#include <rsys/double3.h>
#include <rsys/mem_allocator.h>
#include <rsys/ref_count.h>

#include <string.h>

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static INLINE int
check_bxdf_type(const struct ssf_bxdf_type* type)
{
  return type
      && type->init
      && type->release
      && type->sample
      && type->eval
      && type->pdf
      && IS_POW2(type->alignof_bxdf);
}

static void
bxdf_release(ref_T* ref)
{
  struct ssf_bxdf* bxdf = CONTAINER_OF(ref, struct ssf_bxdf, ref);
  ASSERT(ref);
  if(bxdf->data) {
    bxdf->type.release(bxdf->data);
    MEM_RM(bxdf->allocator, bxdf->data);
  }
  MEM_RM(bxdf->allocator, bxdf);
}

/*******************************************************************************
 * Exported functions
 ******************************************************************************/
res_T
ssf_bxdf_create
  (struct mem_allocator* allocator,
   const struct ssf_bxdf_type* type,
   struct ssf_bxdf** out_bxdf)
{
  struct mem_allocator* mem_allocator = NULL;
  struct ssf_bxdf* bxdf = NULL;
  res_T res = RES_OK;

  if(!out_bxdf || !check_bxdf_type(type)) {
    res = RES_BAD_ARG;
    goto error;
  }
  mem_allocator = allocator ? allocator : &mem_default_allocator;
  bxdf = MEM_CALLOC(mem_allocator, 1, sizeof(struct ssf_bxdf));
  if(!bxdf) {
    res = RES_MEM_ERR;
    goto error;
  }
  ref_init(&bxdf->ref);
  bxdf->allocator = mem_allocator;
  bxdf->type = *type;

  bxdf->data = MEM_ALLOC_ALIGNED
    (bxdf->allocator, bxdf->type.sizeof_bxdf, bxdf->type.alignof_bxdf);
  if(!bxdf->data) {
    res = RES_MEM_ERR;
    goto error;
  }
  memset(bxdf->data, 0, bxdf->type.sizeof_bxdf);
  res = bxdf->type.init(mem_allocator, bxdf->data);
  if(res != RES_OK) goto error;

exit:
  if(out_bxdf) *out_bxdf = bxdf;
  return res;
error:
  if(bxdf) {
    SSF(bxdf_ref_put(bxdf));
    bxdf = NULL;
  }
  goto exit;
}

res_T
ssf_bxdf_ref_get(struct ssf_bxdf* bxdf)
{
  if(!bxdf) return RES_BAD_ARG;
  ref_get(&bxdf->ref);
  return RES_OK;
}

res_T
ssf_bxdf_ref_put(struct ssf_bxdf* bxdf)
{
  if(!bxdf) return RES_BAD_ARG;
  ref_put(&bxdf->ref, bxdf_release);
  return RES_OK;
}

double
ssf_bxdf_sample
  (struct ssf_bxdf* bxdf,
   struct ssp_rng* rng,
   const double wo[3],
   const double N[3],
   double wi[3],
   int* type,
   double* pdf)
{
  ASSERT(bxdf && rng && wo && N && wi && type && pdf);
  ASSERT(d3_is_normalized(wo) && d3_is_normalized(N));
  return bxdf->type.sample(bxdf->data, rng, wo, N, wi, type, pdf);
}

double
ssf_bxdf_eval
  (struct ssf_bxdf* bxdf,
   const double wo[3],
   const double N[3],
   const double wi[3])
{
  ASSERT(bxdf && wo && N && wi);
  ASSERT(d3_is_normalized(wo) && d3_is_normalized(N) && d3_is_normalized(wi));
  return bxdf->type.eval(bxdf->data, wo, N, wi);
}

double
ssf_bxdf_pdf
  (struct ssf_bxdf* bxdf,
   const double wo[3],
   const double N[3],
   const double wi[3])
{
  ASSERT(bxdf && wi && wo);
  ASSERT(d3_is_normalized(wo) && d3_is_normalized(N) && d3_is_normalized(wi));
  return bxdf->type.pdf(bxdf->data, wo, N, wi);
}

res_T
ssf_bxdf_get_data(struct ssf_bxdf* bxdf, void** data)
{
  if(!bxdf || !data) return RES_BAD_ARG;
  *data = bxdf->data;
  return RES_OK;
}

